# -*- coding: utf-8 -*-
"""
"""
import sys
import types
from collections import Counter
from uuid import uuid4

import pytest

from uhashring import HashRing

PY3 = sys.version_info >= (3,)


@pytest.fixture
def ring():
    ring = HashRing(nodes={"node1": 1, "node2": 1, "node3": 1}, hash_fn="ketama")
    return ring


@pytest.fixture
def ring_fast():
    ring = HashRing(nodes={"node1": 1, "node2": 1, "node3": 1}, compat=False)
    return ring


def weight_fn(**conf):
    print(conf)
    return int(conf["nodename"][-1])


def test_ring_size(ring, ring_fast):
    assert ring.size == len(ring.get_nodes()) * 160
    assert ring.size == len(ring._keys)
    assert ring.size == len(ring._ring)
    assert ring.size == len(ring.get_points())

    assert ring_fast.size == len(ring_fast.get_nodes()) * 160
    assert ring_fast.size == len(ring_fast._keys)
    assert ring_fast.size == len(ring_fast._ring)
    assert ring_fast.size == len(ring_fast.get_points())

    assert ring.size == ring_fast.size


def test_consistency(ring):
    coconut_node = ring.get("coconut")
    zoro_node = ring.get("zoro")

    del ring["node3"]

    with pytest.raises(KeyError):
        del ring["node3"]

    assert coconut_node == ring.get("coconut")
    assert zoro_node != ring.get("zoro")

    ring.add_node("node3", 1)

    assert zoro_node == ring.get("zoro")


def test_aliases(ring):
    assert ring.add_node == ring.__setitem__
    assert ring.get_node_instance == ring.__getitem__
    assert ring.remove_node == ring.__delitem__
    assert ring.continuum == ring.ring
    assert ring.nodes == ring.conf


def test_ketama_ring_hashi(ring):
    h = ring.hashi("test")
    assert h == 3446378249 == ring.get_key("test")


def test_faster_ring_hashi(ring_fast):
    h = ring_fast.hashi("test")
    assert h == 12707736894140473154801792860916528374


# XXX https://docs.python.org/3/whatsnew/3.0.html#ordering-comparisons
@pytest.mark.skipif(PY3, reason="requires python26, python27")
def test_range(ring):
    r = list(ring.range("test"))
    r.sort()
    c = ring.conf.values()
    c.sort()
    assert r == c

    r = list(ring.range("test", size=2, unique=True))
    assert len(r) == 2

    r = list(ring.range("test", size=2, unique=False))
    assert len(r) == 2

    r = list(ring.range("1800"))  # 1800 is at position 0
    r = list(ring.range("849"))  # 849 is at position -2

    r = list(ring.range("test", size=None, unique=False))
    assert len(r) == ring.size

    n = []
    for node in ring.iterate_nodes("test"):
        n.append(node)
    assert len(n) == 3

    ring_empty = HashRing()
    for node in ring_empty.iterate_nodes("test"):
        assert node is None


def test_conf():
    ring_1 = HashRing({"node1": 1})
    ring_2 = HashRing("node1")
    ring_3 = HashRing(["node1"])
    ring_4 = HashRing({"node1": {"weight": 1}})

    assert ring_1.ring == ring_2.ring == ring_3.ring == ring_4.ring

    with pytest.raises(ValueError):
        HashRing({"node1": "fail"})

    with pytest.raises(ValueError):
        HashRing(None)


def test_empty_ring():
    ring = HashRing()
    assert ring.get("test") is None


def test_methods_return_types(ring):
    assert isinstance(ring["test"], type(None))
    assert isinstance(ring.get("test"), dict)
    assert isinstance(ring.get_instances(), list)
    assert isinstance(ring.get_node("test"), str)
    assert isinstance(ring.get_nodes(), type({}.keys()) if PY3 else list)
    assert isinstance(ring.get_node_hostname("test"), str)
    assert isinstance(ring.get_node_port("test"), type(None))
    assert isinstance(ring.get_node_pos("test"), int)
    assert isinstance(ring.get_node_weight("test"), int)
    assert isinstance(ring.get_points(), list)
    assert isinstance(ring.get_server("test"), tuple)
    assert isinstance(ring.iterate_nodes("test"), types.GeneratorType)
    assert isinstance(ring.range("test"), types.GeneratorType)
    assert isinstance(ring.distribution, Counter)


def test_print_without_error(ring):
    assert ring.print_continuum() is None
    ring = HashRing()
    assert ring.print_continuum() is None


def test_with_non_str_objects(ring):
    uid = uuid4()
    ring.get_node(uid)


def test_weight_fn():
    ring = HashRing(
        nodes={"node1": 1, "node2": 1, "node3": 1},
        replicas=4,
        vnodes=40,
        hash_fn="ketama",
        weight_fn=weight_fn,
    )

    assert ring.distribution["node1"] == 80
    assert ring.distribution["node2"] == 160
    assert ring.distribution["node3"] == 240

    ring.regenerate

    assert ring.distribution["node1"] == 80
    assert ring.distribution["node2"] == 160
    assert ring.distribution["node3"] == 240

    with pytest.raises(TypeError):
        ring = HashRing(
            nodes={"node1": 1, "node2": 1, "node3": 1},
            replicas=4,
            vnodes=40,
            hash_fn="ketama",
            weight_fn=12,
        )

    with pytest.raises(TypeError):
        ring = HashRing(
            nodes={"node1": 1, "node2": 1, "node3": 1},
            replicas=4,
            vnodes=40,
            hash_fn="ketama",
            weight_fn="coconut",
        )


def test_ring_growth_ketama(ring):
    add_ring = HashRing(hash_fn="ketama")
    for nodename in ring.nodes:
        add_ring.add_node(nodename)

    assert ring._nodes == add_ring._nodes
    assert ring.ring == add_ring.ring
    assert ring.distribution == add_ring.distribution


def test_ring_growth_meta(ring_fast):
    add_ring = HashRing(compat=False)
    for nodename in ring_fast.nodes:
        add_ring.add_node(nodename)

    assert ring_fast._nodes == add_ring._nodes
    assert ring_fast.ring == add_ring.ring
    assert ring_fast.distribution == add_ring.distribution


def test_ketama_ring_shrink_collision():
    """
    see issue #6 thanks to @bjhockley
    """
    nodes = ["172.31.1.0", "172.31.1.125", "172.31.1.202"]
    ring = HashRing(nodes, hash_fn="ketama")
    ring.remove_node(nodes[1])
    ring.remove_node(nodes[2])
    ring.remove_node(nodes[0])
    assert ring.ring == {}


def test_hash_fn():
    """"""

    def hash_fn(k):
        return k + "_hash"

    nodes = ["172.31.1.0", "172.31.1.125", "172.31.1.202"]
    ring = HashRing(nodes, hash_fn=hash_fn)
    assert ring.hashi("coconut") == "coconut_hash"


def test_ketama_replica_arg():
    ring = HashRing([1, 2, 3, 4], hash_fn="ketama", replicas=3)
    assert ring.runtime._replicas == 3
    assert ring.get_node("foo") == 4



from bisect import bisect

from uhashring.ring_ketama import KetamaRing
from uhashring.ring_meta import MetaRing


class HashRing:
    """Implement a consistent hashing ring."""

    def __init__(self, nodes=[], **kwargs):
        """Create a new HashRing given the implementation.

        :param nodes: nodes used to create the continuum (see doc for format).
        :param hash_fn: use this callable function to hash keys, can be set to
                        'ketama' to use the ketama compatible implementation.
        :param vnodes: default number of vnodes per node.
        :param weight_fn: use this function to calculate the node's weight.
        """
        hash_fn = kwargs.get("hash_fn", None)
        vnodes = kwargs.get("vnodes", None)
        weight_fn = kwargs.get("weight_fn", None)

        if hash_fn == "ketama":
            ketama_args = {k: v for k, v in kwargs.items() if k in ("replicas",)}
            if vnodes is None:
                vnodes = 40
            self.runtime = KetamaRing(**ketama_args)
        else:
            if vnodes is None:
                vnodes = 160
            self.runtime = MetaRing(hash_fn)

        self._default_vnodes = vnodes
        self.hashi = self.runtime.hashi

        if weight_fn and not hasattr(weight_fn, "__call__"):
            raise TypeError("weight_fn should be a callable function")
        self._weight_fn = weight_fn

        if self._configure_nodes(nodes):
            self.runtime._create_ring(self.runtime._nodes.items())

    def _configure_nodes(self, nodes):
        """Parse and set up the given nodes.

        :param nodes: nodes used to create the continuum (see doc for format).
        """
        if isinstance(nodes, str):
            nodes = [nodes]
        elif not isinstance(nodes, (dict, list)):
            raise ValueError(
                "nodes configuration should be a list or a dict," " got {}".format(type(nodes))
            )

        conf_changed = False
        for node in nodes:
            conf = {
                "hostname": node,
                "instance": None,
                "nodename": node,
                "port": None,
                "vnodes": self._default_vnodes,
                "weight": 1,
            }
            current_conf = self.runtime._nodes.get(node, {})
            nodename = node
            # new node, trigger a ring update
            if not current_conf:
                conf_changed = True
            # complex config
            if isinstance(nodes, dict):
                node_conf = nodes[node]
                if isinstance(node_conf, int):
                    conf["weight"] = node_conf
                elif isinstance(node_conf, dict):
                    for k, v in node_conf.items():
                        if k in conf:
                            conf[k] = v
                            # changing those config trigger a ring update
                            if k in ["nodename", "vnodes", "weight"]:
                                if current_conf.get(k) != v:
                                    conf_changed = True
                else:
                    raise ValueError(
                        "node configuration should be a dict or an int,"
                        " got {}".format(type(node_conf))
                    )
            if self._weight_fn:
                conf["weight"] = self._weight_fn(**conf)
            # changing the weight of a node trigger a ring update
            if current_conf.get("weight") != conf["weight"]:
                conf_changed = True
            self.runtime._nodes[nodename] = conf
        return conf_changed

    def __delitem__(self, nodename):
        """Remove the given node.

        :param nodename: the node name.
        """
        self.runtime._remove_node(nodename)

    remove_node = __delitem__

    def __getitem__(self, key):
        """Returns the instance of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "instance")

    get_node_instance = __getitem__

    def __setitem__(self, nodename, conf={"weight": 1}):
        """Add the given node with its associated configuration.

        :param nodename: the node name.
        :param conf: the node configuration.
        """
        if self._configure_nodes({nodename: conf}):
            self.runtime._create_ring([(nodename, self._nodes[nodename])])

    add_node = __setitem__

    def _get_pos(self, key):
        """Get the index of the given key in the sorted key list.

        We return the position with the nearest hash based on
        the provided key unless we reach the end of the continuum/ring
        in which case we return the 0 (beginning) index position.

        :param key: the key to hash and look for.
        """
        p = bisect(self.runtime._keys, self.hashi(key))
        if p == len(self.runtime._keys):
            return 0
        else:
            return p

    def _get(self, key, what):
        """Generic getter magic method.

        The node with the nearest but not less hash value is returned.

        :param key: the key to look for.
        :param what: the information to look for in, allowed values:
            - instance (default): associated node instance
            - nodename: node name
            - pos: index of the given key in the ring
            - tuple: ketama compatible (pos, name) tuple
            - weight: node weight
        """
        if not self.runtime._ring:
            return None

        pos = self._get_pos(key)
        if what == "pos":
            return pos

        nodename = self.runtime._ring[self.runtime._keys[pos]]
        if what in ["hostname", "instance", "port", "weight"]:
            return self.runtime._nodes[nodename][what]
        elif what == "dict":
            return self.runtime._nodes[nodename]
        elif what == "nodename":
            return nodename
        elif what == "tuple":
            return (self.runtime._keys[pos], nodename)

    def get(self, key):
        """Returns the node object dict matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "dict")

    def get_instances(self):
        """Returns a list of the instances of all the configured nodes."""
        return [c.get("instance") for c in self.runtime._nodes.values() if c.get("instance")]

    def get_key(self, key):
        """Alias of ketama hashi method, returns the hash of the given key.

        This method is present for hash_ring compatibility.

        :param key: the key to look for.
        """
        return self.hashi(key)

    def get_node(self, key):
        """Returns the node name of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "nodename")

    def get_node_hostname(self, key):
        """Returns the hostname of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "hostname")

    def get_node_port(self, key):
        """Returns the port of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "port")

    def get_node_pos(self, key):
        """Returns the index position of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "pos")

    def get_node_weight(self, key):
        """Returns the weight of the node matching the hashed key.

        :param key: the key to look for.
        """
        return self._get(key, "weight")

    def get_nodes(self):
        """Returns a list of the names of all the configured nodes."""
        return self.runtime._nodes.keys()

    def get_points(self):
        """Returns a ketama compatible list of (position, nodename) tuples."""
        return [(k, self.runtime._ring[k]) for k in self.runtime._keys]

    def get_server(self, key):
        """Returns a ketama compatible (position, nodename) tuple.

        :param key: the key to look for.
        """
        return self._get(key, "tuple")

    def iterate_nodes(self, key, distinct=True):
        """hash_ring compatibility implementation.

        Given a string key it returns the nodes as a generator that
        can hold the key.
        The generator iterates one time through the ring
        starting at the correct position.
        if `distinct` is set, then the nodes returned will be unique,
        i.e. no virtual copies will be returned.
        """
        if not self.runtime._ring:
            yield None
        else:
            for node in self.range(key, unique=distinct):
                yield node["nodename"]

    def print_continuum(self):
        """Prints a ketama compatible continuum report."""
        numpoints = len(self.runtime._keys)
        if numpoints:
            print(f"Numpoints in continuum: {numpoints}")
        else:
            print("Continuum empty")
        for p in self.get_points():
            point, node = p
            print(f"{node} ({point})")

    def range(self, key, size=None, unique=True):
        """Returns a generator of nodes' configuration available
        in the continuum/ring.

        :param key: the key to look for.
        :param size: limit the list to at most this number of nodes.
        :param unique: a node may only appear once in the list (default True).
        """
        all_nodes = set()
        if unique:
            size = size or len(self.runtime._nodes)
        else:
            all_nodes = []

        pos = self._get_pos(key)
        for key in self.runtime._keys[pos:]:
            nodename = self.runtime._ring[key]
            if unique:
                if nodename in all_nodes:
                    continue
                all_nodes.add(nodename)
            else:
                all_nodes.append(nodename)
            yield self.runtime._nodes[nodename]
            if len(all_nodes) == size:
                break
        else:
            for i, key in enumerate(self.runtime._keys):
                if i < pos:
                    nodename = self.runtime._ring[key]
                    if unique:
                        if nodename in all_nodes:
                            continue
                        all_nodes.add(nodename)
                    else:
                        all_nodes.append(nodename)
                    yield self.runtime._nodes[nodename]
                    if len(all_nodes) == size:
                        break

    def regenerate(self):
        self.runtime._create_ring(self.runtime._nodes.items())

    @property
    def conf(self):
        return self.runtime._nodes

    nodes = conf

    @property
    def distribution(self):
        return self.runtime._distribution

    @property
    def ring(self):
        return self.runtime._ring

    continuum = ring

    @property
    def size(self):
        return len(self.runtime._ring)

    @property
    def _ring(self):
        return self.runtime._ring

    @property
    def _nodes(self):
        return self.runtime._nodes

    @property
    def _keys(self):
        return self.runtime._keys



from uhashring import HashRing

__all__ = ["patch_memcache"]


def patch_memcache():
    """Monkey patch python-memcached to implement our consistent hashring
    in its node selection and operations.
    """

    def _init(self, servers, *k, **kw):
        self._old_init(servers, *k, **kw)

        nodes = {}
        for server in self.servers:
            conf = {
                "hostname": server.ip,
                "instance": server,
                "port": server.port,
                "weight": server.weight,
            }
            nodes[server.ip] = conf
        self.uhashring = HashRing(nodes)

    def _get_server(self, key):
        if isinstance(key, tuple):
            return self._old_get_server(key)

        for i in range(self._SERVER_RETRIES):
            for node in self.uhashring.range(key):
                if node["instance"].connect():
                    return node["instance"], key

        return None, None

    memcache = __import__("memcache")
    memcache.Client._old_get_server = memcache.Client._get_server
    memcache.Client._old_init = memcache.Client.__init__
    memcache.Client.__init__ = _init
    memcache.Client._get_server = _get_server



from bisect import insort
from collections import Counter
from hashlib import md5


class KetamaRing:
    """Implement a ketama compatible consistent hashing ring."""

    def __init__(self, replicas=4):
        """Create a new HashRing."""
        self._distribution = Counter()
        self._keys = []
        self._nodes = {}
        self._replicas = replicas
        self._ring = {}

        self._listbytes = lambda x: x

    def hashi(self, key, replica=0):
        """Returns a ketama compatible hash from the given key."""
        dh = self._listbytes(md5(str(key).encode("utf-8")).digest())
        rd = replica * 4
        return (dh[3 + rd] << 24) | (dh[2 + rd] << 16) | (dh[1 + rd] << 8) | dh[0 + rd]

    def _hashi_weight_generator(self, node_name, node_conf):
        """Calculate the weight factor of the given node and
        yield its hash key for every configured replica.

        :param node_name: the node name.
        """
        ks = (node_conf["vnodes"] * len(self._nodes) * node_conf["weight"]) // self._weight_sum
        for w in range(0, ks):
            w_node_name = f"{node_name}-{w}"
            for i in range(0, self._replicas):
                yield self.hashi(w_node_name, replica=i)

    @staticmethod
    def _listbytes(data):
        """Python 2 compatible int iterator from str.

        :param data: the string to int iterate upon.
        """
        return map(ord, data)

    def _create_ring(self, nodes):
        """Generate a ketama compatible continuum/ring."""
        _weight_sum = 0
        for node_conf in self._nodes.values():
            _weight_sum += node_conf["weight"]
        self._weight_sum = _weight_sum

        _distribution = Counter()
        _keys = []
        _ring = {}
        for node_name, node_conf in self._nodes.items():
            for h in self._hashi_weight_generator(node_name, node_conf):
                _ring[h] = node_name
                insort(_keys, h)
                _distribution[node_name] += 1
        self._distribution = _distribution
        self._keys = _keys
        self._ring = _ring

    def _remove_node(self, node_name):
        """Remove the given node from the continuum/ring.

        :param node_name: the node name.
        """
        try:
            self._nodes.pop(node_name)
        except Exception:
            raise KeyError(
                "node '{}' not found, available nodes: {}".format(node_name, self._nodes.keys())
            )
        else:
            self._create_ring(self._nodes)



# -*- coding: utf-8 -*-
"""
"""

from uhashring import HashRing, monkey


def test_patch_memcache():
    """Returns if no memcache.

    NB:
        test key -> 11212
        zzzzzzzzzz key -> 11211
    """
    monkey.patch_memcache()

    import memcache

    mc = memcache.Client(["127.0.0.1:11211", "127.0.0.0:11212"])
    mc.set("zzzzzzzzzz", 1, 2)
    mc.get("zzzzzzzzzz")
    mc.get("test")
    mc.get((0, "zzzzzzzzzz"))

    assert isinstance(mc.uhashring, HashRing)



# -*- coding: utf-8 -*-
"""This is not part of the test suite.
"""
try:
    import ketama
except Exception:
    ketama = None

from tempfile import NamedTemporaryFile
from time import time

from uhashring import HashRing

num = 1000000
print("running {} key generation comparison".format(num))

# ketama C binding
if ketama:
    with NamedTemporaryFile(prefix="benchmark_") as ketama_config_file:
        ketama_config_file.write("127.0.0.1:11211\t600\n")
        ketama_config_file.write("127.0.0.1:11212\t400\n")
        ketama_config_file.flush()

        kt = ketama.Continuum(ketama_config_file.name)
        pt = time()
        for i in range(num):
            key = "myval-{}".format(i)
            kt.get_server(key)
        print("ketama took {} s".format(time() - pt))

# pure python implementation
ring = HashRing(
    nodes={"127.0.0.1:11211": 600, "127.0.0.1:11212": 400},
    replicas=4,
    vnodes=40,
    compat=True,
)
pt = time()
for i in range(num):
    key = "myval-{}".format(i)
    ring.get_server(key)
print("HashRing took {} s".format(time() - pt))



# -*- coding: utf-8 -*-
"""This test is vastly inspired by Mike Bayer's article:
http://techspot.zzzeek.org/2012/07/07/the-absolutely-simplest-consistent-hashing-example/
"""

from collections import Counter
from itertools import cycle
from math import sqrt

from uhashring import HashRing


def _pop_std_dev(population):
    mean = sum(population) / len(population)
    return sqrt(sum(pow(n - mean, 2) for n in population) / len(population))


def test_distribution():
    ring = HashRing()

    numnodes = 10
    numhits = 1000
    numvalues = 10000

    for i in range(1, numnodes + 1):
        ring["node{}".format(i)] = {"instance": "node_value{}".format(i)}

    deterministic_input = cycle(range(1, numvalues))
    distribution = Counter()
    for i in range(numhits):
        key = str(next(deterministic_input))
        node = ring[key]
        distribution[node] += 1

    # count of hits matches what is observed
    assert sum(distribution.values()) == numhits

    # usually under 20
    standard_dev = _pop_std_dev(distribution.values())
    assert standard_dev <= 20

    # all nodes should be used
    assert len(distribution) == numnodes

    # just to test getting keys, see that we got the values
    # back and not keys or indexes or whatever.
    assert set(distribution.keys()) == set("node_value{}".format(i) for i in range(1, 1 + numnodes))



# -*- coding: utf-8 -*-
"""ketama is not released under pypi !
This test is only run on my dev machine and is not really part of the CI.
"""
try:
    import ketama
except Exception:
    ketama = None
from random import randint
from tempfile import NamedTemporaryFile

import pytest

from uhashring import HashRing


@pytest.fixture
def ketama_config_file(request):
    valid_list = NamedTemporaryFile(prefix="py.test_")
    valid_list.write(b"127.0.0.1:11211\t600\n")
    valid_list.write(b"127.0.0.1:11212\t400\n")
    valid_list.flush()

    def fin():
        valid_list.close()
        print("closed valid_list")

    request.addfinalizer(fin)
    return valid_list.name


def test_ketama_hashi():
    if not ketama:
        return

    ring = HashRing()
    assert ring.hashi("test") == ketama.hashi("test")


def test_ketama_compatibility(ketama_config_file):
    if not ketama:
        return

    ring = HashRing(
        nodes={"127.0.0.1:11211": 600, "127.0.0.1:11212": 400},
        replicas=4,
        vnodes=40,
        compat=True,
    )
    continuum = ketama.Continuum(ketama_config_file)

    assert ring.get_points() == continuum.get_points()

    numhits = 1000
    numvalues = 10000
    for i in range(numhits):
        key = str(randint(1, numvalues))
        assert ring.get_server(key) == continuum.get_server(key)



# -*- coding: utf-8 -*-
"""This is not part of the test suite.
"""
from hashlib import md5, sha1, sha256
from time import time

num = 1000000
print("running {} key generation comparison".format(num))

# md5
pt = time()
for i in range(num):
    key = b"myval-%d" % i
    int(md5(b"%s" % key).hexdigest(), 16)
print("md5 took {} s".format(time() - pt))

# sha1
pt = time()
for i in range(num):
    key = b"myval-%d" % i
    int(sha1(b"%s" % key).hexdigest(), 16)
print("sha1 took {} s".format(time() - pt))

# sha256
pt = time()
for i in range(num):
    key = b"myval-%d" % i
    int(sha256(b"%s" % key).hexdigest(), 16)
print("sha256 took {} s".format(time() - pt))

# CityHash
try:
    from cityhash import CityHash32, CityHash64, CityHash128
except ImportError:
    print("CityHash not installed, pip install cityhash")
else:
    # cityhash32
    pt = time()
    for i in range(num):
        key = b"myval-%d" % i
        CityHash32(b"%s" % key)
    print("CityHash32 took {} s".format(time() - pt))

    # cityhash64
    pt = time()
    for i in range(num):
        key = b"myval-%d" % i
        CityHash64(b"%s" % key)
    print("CityHash64 took {} s".format(time() - pt))

    # cityhash128
    pt = time()
    for i in range(num):
        key = b"myval-%d" % i
        CityHash128(b"%s" % key)
    print("CityHash128 took {} s".format(time() - pt))

# murmurhash v3
try:
    import mmh3
except ImportError:
    print("MurmurHash v3 not installed, pip install murmurhash3")
else:
    pt = time()
    for i in range(num):
        key = b"myval-%d" % i
        mmh3.hash(b"%s" % key)
    print("MurmurHash v3 took {} s".format(time() - pt))
